from dataclasses import dataclass, field
import datetime as dt
import asyncio
import importlib
import json
import uuid
import time
from Library.utils.DictDot import DictDot

from Library.DomoClasses.routes import publish_routes

import Library.DomoClasses.DomoDataset as dmda
import Library.DomoClasses.DomoLineage as dmdl

importlib.reload(dmdl)

# class InvalidUrl


@dataclass
class DomoPublication_Subscription:
    subscription_id: str
    domain: str
    created_dt: dt.datetime

@dataclass
class DomoPublication_Content:
    content_id: str
    entity_type: str
    entity_id: str
    content_domain: str
    
    def to_json (self):
        temp_dict = {'domain': self.content_domain,
                 'domoObjectId': self.entity_id,
                 'customerId': self.content_domain,
                 'type': self.entity_type}
        return temp_dict


@dataclass
class DomoPublication:
    id: str
    name: str
    description: str
    is_v2: bool
    created_dt: dt.datetime
    full_auth: dmda.DomoFullAuth = field(default=None, repr=False)

    subscription_authorizations: [
        DomoPublication_Subscription] = field(default_factory=list)
    content_entity_ls: [DomoPublication_Content] = field(default_factory=list)

    content_page_id_ls: [str] = None
    content_dataset_id_ls: [str] = None

    lineage: dmdl.DomoLineage = None

    def __post_init__(self):
        self.lineage = dmdl.DomoLineage(id=self.id,
                                        parent=self)

    @classmethod
    def _from_json(cls, obj, full_auth: dmda.DomoFullAuth = None):
        dd = DictDot(obj)

        domo_pub = cls(
            id=dd.id,
            name=dd.name,
            description=dd.description,
            created_dt=dt.datetime.fromtimestamp(
                dd.created/1000) if dd.created else None,
            is_v2=dd.isV2,
            full_auth=full_auth
        )

        if dd.subscription_authorizations and len(dd.subscription_authorizations) > 0:
            domo_pub.subscription_authorizations = [DomoPublication_Subscription(subscription_id=sub.id,
                                                                                 domain=sub.domain,
                                                                                 created_dt=dt.datetime.fromtimestamp(
                                                                                     sub.created/1000)
                                                                                 if sub.created else None
                                                                                 )
                                                    for sub in dd.subscription_authorizations]

        # publish only supports sharing pages and datasets
        if dd.children and len(dd.children) > 0:
            for child in dd.children:
                dmpc = DomoPublication_Content(
                    content_id=child.id,
                    entity_type=child.content.type,
                    entity_id=child.content.domoObjectId,
                    content_domain=child.content.domain)

                if not domo_pub.content_entity_ls:
                    domo_pub.content_entity_ls = []

                domo_pub.content_entity_ls.append(dmpc)

                if dmpc.entity_type == 'PAGE':
                    if not domo_pub.content_page_id_ls:
                        domo_pub.content_page_id_ls = []
                    domo_pub.content_page_id_ls.append(dmpc.entity_id)

                if dmpc.entity_type == 'DATASET':
                    if not domo_pub.content_dataset_id_ls:
                        domo_pub.content_dataset_id_ls = []
                    domo_pub.content_dataset_id_ls.append(dmpc.entity_id)

        return domo_pub

    @classmethod
    async def get_from_id(cls, publication_id=None, full_auth: dmda.DomoFullAuth = None):

        full_auth = full_auth or cls.full_auth

        publication_id = publication_id or cls.publication_id

        res = await publish_routes.get_publication_by_id(full_auth=full_auth, publication_id=publication_id)

        if res.status != 200:
            print(res)
            return None

        return cls._from_json(obj=res.response, full_auth=full_auth)

    def convert_content_to_dataframe(self, return_raw: bool = False):

        output_ls = [{'plubication_id': self.id,
                      'publication_name': self.name,
                      'is_v2': self.is_v2,
                      'publish_created_dt': self.created_dt,
                      'entity_type': row.type,
                      'entity_id': row.id
                      } for row in self.content_entity_ls]

        if return_raw:
            return output_ls

        return pd.DataFrame(output_ls)

    def convert_lineage_to_dataframe(self, return_raw: bool = False):
        import pandas as pd
        import re

        flat_lineage_ls = self.lineage._flatten_lineage()

        output_ls = [{'plubication_id': self.id,
                      'publication_name': self.name,
                      'is_v2': self.is_v2,
                      'publish_created_dt': self.created_dt,
                      'entity_type': row.get('entity_type'),
                      'entity_id': row.get('entity_id')
                      } for row in flat_lineage_ls]

        if return_raw:
            return output_ls

        return pd.DataFrame(output_ls)
    
    @classmethod
    async def create_publication(cls, 
                                 name : str, 
                                 content_ls : [DomoPublication_Content], 
                                 subscription_ls : [DomoPublication_Subscription],
                                 unique_id : str = None,
                                 description: str = None,
                                 full_auth: dmda.DomoFullAuth = None,
                                 debug: bool = False):
        
        if not isinstance(subscription_ls, list) :
            subscription_ls = [subscription_ls]

        full_auth = full_auth or cls.full_auth
        domain_ls =[]
        content_json_ls =[]
        for sub in subscription_ls:
            domain_ls.append(sub.domain)
        for content_item in content_ls:
            content_json_ls.append(content_item.to_json())
        
        if not unique_id:
            unique_id = str(uuid.uuid4())
        if not description:
            description =''
          
        body = publish_routes.generate_publish_body(url = f'{full_auth.domo_instance}.domo.com',
                                                    sub_domain_ls= domain_ls,
                                                    content_ls = content_json_ls,
                                                    name = name,
                                                    unique_id=unique_id,
                                                    description= description,
                                                    is_new  = True)
        
        res = await publish_routes.create_publish_job(full_auth = full_auth, body=body)
        if debug:
            print('Create the new Publish job')
        if res.status != 200:
            print(res)
            await asyncio.sleep(2) 
            res = await publish_routes.get_publication_by_id(full_auth=full_auth, publication_id=unique_id)
            if res.status != 200:
                return None
            else:
                return cls._from_json(obj=res.response, full_auth=full_auth)

        return cls._from_json(obj=res.response, full_auth=full_auth)

    
    @classmethod
    async def update_publication(cls, 
                                 name : str, 
                                 content_ls : [DomoPublication_Content], 
                                 subscription_ls : [DomoPublication_Subscription],
                                 publication_id : str,
                                 description: str = None, 
                                 full_auth: dmda.DomoFullAuth = None,
                                 debug: bool = False):
        
        if not isinstance(subscription_ls, list) :
            subscription_ls = [subscription_ls]

        full_auth = full_auth or cls.full_auth
        domain_ls =[]
        content_json_ls =[]
        for sub in subscription_ls:
            domain_ls.append(sub.domain)
        for content_item in content_ls:
            content_json_ls.append(content_item.to_json())

        if not description:
            description =''
        body = publish_routes.generate_publish_body(url = f'{full_auth.domo_instance}.domo.com',
                                                    sub_domain_ls= domain_ls,
                                                    content_ls = content_json_ls,
                                                    name = name,
                                                    unique_id=publication_id,
                                                    description= description,
                                                    is_new  = False)
        
        res = await publish_routes.udpate_publish_job(full_auth = full_auth, 
                                                      publication_id = publication_id,
                                                      body = body)
        if debug:
            print('Update Publish job by id')
        if res.status != 200:
            print(res)
            await asyncio.sleep(2) 
            res = await publish_routes.get_publication_by_id(full_auth=full_auth, publication_id=publication_id)
            if res.status != 200:
                return None
            else:
                return cls._from_json(obj=res.response, full_auth=full_auth)

        return cls._from_json(obj=res.response, full_auth=full_auth)
    
    @classmethod
    async def get_subscription_invites_list(cls, full_auth: dmda.DomoFullAuth,
                                            debug: bool = False):
        
        res = await publish_routes.get_subscription_invites_list(full_auth=full_auth,
                                                       debug=debug)
        if debug:
            print('Getting Publish subscription invites')

        if res.status == 200:
            return res.response
    
    @classmethod
    async def accept_invite_by_id(cls,
                                    full_auth: dmda.DomoFullAuth, 
                                    subscription_id : str,
                                    debug: bool = False):
        
        res = await publish_routes.accept_invite_by_id(full_auth=full_auth,
                                                        subscription_id=subscription_id,
                                                       debug=debug)
        if debug:
            print(f'Accept invite by id {subscription_id}')

        if res.status == 200:
            return res.response
    


from typing import Any, Awaitable


async def run_sequence(*functions: Awaitable[Any]) -> None:
    for function in functions:
        await function


def chunk_list(tlist, chunk_size):
    chunked_list = [tlist[i * chunk_size:(i + 1) * chunk_size] for i in
                    range((len(tlist) + chunk_size - 1) // chunk_size)]
    return chunked_list


async def chunk_fn(chunk, api_fun, idx, sleep_time, session, list_id):
    print(f'sleeping {idx} and {len(chunk)}')

    res = await asyncio.gather(*[api_fun(row=row, session=session, list_id=list_id) for row in chunk])

    await asyncio.sleep(sleep_time)
    print(f'end_sleep {idx}')
    return res


async def api_request_in_chunks(full_list, api_fn, api_limit_size, list_id, sleep_time=10):
    chunked_list = chunk_list(tlist=full_list, chunk_size=api_limit_size)
    session = aiohttp.ClientSession(request_class=OAuthRequest)
    res = await run_sequence(
        *[chunk_fn(chunk, api_fn, idx, sleep_time=sleep_time, session=session, list_id=list_id) for idx, chunk in
          enumerate(chunked_list)])
    await session.close()
    return res



from dataclasses import dataclass, field

import aiohttp

from ..utils import Exceptions as ex
from ..utils import ResponseGetData as gd
from .routes import auth_routes


class Error(Exception):
    """Base class for other exceptions"""
    pass


class InvalidCredentialsError(Exception):
    def __init__(self, status, message="invalid credentials", domo_instance=None):

        instance_str = f" at {domo_instance}" if domo_instance else None
        self.message = f"Status {status} - {message}{instance_str or ''}" or message
        super().__init__(self.message)


@dataclass
class _DA_Base:
    domo_instance: str


@dataclass
class _DA_Default:
    token: str = field(default=None, repr=False)
    token_name: str = field(default=None)
    user_id: str = field(default=None, repr=False)
    auth_header: dict = field(default_factory=dict, repr=False)

    url_manual_login: str = None

    async def print_is_token(self, token_name=None) -> None:
        self.token_name = token_name
        if not self.token:
            await self.get_auth_token()

        token_str = f"{token_name} "

        if not self.token:
            print(
                f"🚧 failed to retrieve {token_str if token_name else ''}token from {self.domo_instance}")
            return False

        print(
            f"🎉 {token_str if token_name else ''}token retrieved from {self.domo_instance} ⚙️")
        return True


@dataclass
class DomoAuth(_DA_Default, _DA_Base):
    def _init__():
        super().__init__()


@dataclass
class _DFA_Base(_DA_Base):
    domo_username: str
    domo_password: str = field(default=None, repr=False)


@dataclass
class DomoFullAuth(_DA_Default, _DFA_Base):
    """use for full authentication token"""

    def __post_init__(self):
        self.url_manual_login = f"https://{self.domo_instance}.domo.com/auth/index?domoManualLogin=true"

    def generate_auth_header(self, token: str) -> dict:
        self.auth_header = {'x-domo-authentication': token}
        return self.auth_header

    async def get_auth_token(self, domo_instance=None,
                             domo_username=None, domo_password=None,
                             debug: bool = False,
                             session: aiohttp.ClientSession = None) -> gd.ResponseGetData:

        self.domo_username = domo_username or self.domo_username
        self.domo_password = domo_password or self.domo_password
        self.domo_instance = domo_instance or self.domo_instance

        res = await auth_routes.get_full_auth(domo_instance=self.domo_instance,
                                              domo_username=self.domo_username,
                                              domo_password=self.domo_password,
                                              session=session)
        if debug:
            print(res)

        if res.is_success == True:
            self.token = res.response.get('sessionToken')
            self.user_id = res.response.get('userId')

            self.auth_header = self.generate_auth_header(token=self.token)

            return self.token

        elif res.status == 200 and not res.is_success:
            raise InvalidCredentialsError(status=res.status,
                                          message=res.response.get("reason"),
                                          domo_instance=self.domo_instance)

        else:
            raise ex.InvalidInstanceError(
                message=f'invalid instance {self.domo_instance}')


@dataclass(init=False)
class DomoDeveloperAuth(DomoAuth):
    """use for developer authentication token"""
    domo_client_id: str
    domo_client_secret: str
    auth_header: str

    def __init__(self, domo_client_id, domo_client_secret, domo_instance=None):
        self.domo_client_id = domo_client_id
        self.domo_client_secret = domo_client_secret
        super().__init__(domo_instance)

    def generate_auth_header(self, token: str) -> dict:
        self.auth_header = {'Authorization': f'bearer {token}'}
        return self.auth_header

    def generate_auth_header_private(self, token: str) -> dict:
        self.auth_header = {'X-domo-developer-token': f'bearer {token}'}
        return self.auth_header

    async def get_auth_token(self,
                             domo_client_id=None,
                             domo_client_secret=None) -> gd.ResponseGetData:

        self.domo_client_id = domo_client_id or self.domo_client_id
        self.domo_client_secret = domo_client_secret or self.domo_client_secret

        res = await auth_routes.get_developer_auth(domo_client_id=self.domo_client_id,
                                                   domo_client_secret=self.domo_client_secret)

        if res.status == 200 and res.response.get('access_token'):
            self.token = res.response.get('access_token')
            self.user_id = res.response.get('userId')

            self.auth_header = self.generate_auth_header(token=self.token)

            return res

        elif res.status == 200:
            raise InvalidCredentialsError(
                status=res.status, message=res.response.get("reason"))



import asyncio
import datetime as dt
import importlib
import io
import json
from dataclasses import dataclass, field
from enum import Enum, auto
from pprint import pprint
from typing import Any, List

import aiohttp
import pandas as pd

from ..utils import Exceptions as ex
from ..utils.Base import Base
from ..utils.chunk_execution import chunk_list
from ..utils.DictDot import DictDot
from . import DomoCertification as dmdc
from . import DomoPDP as dmpdp
from . import DomoTag as dmtg
from .DomoAuth import DomoDeveloperAuth, DomoFullAuth
from .routes import dataset_routes

importlib.reload(dmtg)


@dataclass
class Schema_Column:
    name: str
    id: str
    type: str
    dataset: Any

    @classmethod
    def _from_json(cls, json_obj, domo_dataset):
        dd = DictDot(json_obj)
        return cls(
            name=dd.name,
            id=dd.id,
            type=dd.type,
            dataset=domo_dataset
        )


@dataclass(init=False)
class Dataset_Schema:
    columns: List[Schema_Column] = field(default_factory=list)

    def __init__(self, dataset):
        self.dataset = dataset
        self.columns = []

    async def get(self, full_auth: DomoFullAuth = None, debug: bool = False) -> List[Schema_Column]:
        full_auth = full_auth or self.dataset.full_auth

        res = await dataset_routes.get_schema(full_auth=self.dataset.full_auth, id=self.dataset.id, debug=debug)

        if res.status == 200:
            json_list = res.response.get('tables')[0].get('columns')

            self.columns = []
            for json_obj in json_list:
                dc = Schema_Column._from_json(
                    json_obj=json_obj, domo_dataset=self.dataset)
                if dc not in self.columns:
                    self.columns.append(dc)

            return self.columns


@dataclass
class DomoDataset:
    "interacts with domo datasets"
    full_auth: DomoFullAuth = field(repr=False, default=None)
    dev_auth: DomoDeveloperAuth = field(repr=False, default=None)

    id: str = ''
    display_type: str = ''
    data_provider_type: str = ''
    name: str = ''
    description: str = ''
    row_count: int = None
    column_count: int = None
    owner: dict = field(default_factory=dict)
    formula: dict = field(default_factory=dict)
    stream_id: int = None
    domo_instance: str = ''

    tags: dmtg.Dataset_Tags = None
    certification: dmdc.DomoCertification = None
    PDPPolicies: dmpdp.Dataset_PDP_Policies = None
    schema: Dataset_Schema = None

    def __post_init__(self):
        self.PDPPolicies = dmpdp.Dataset_PDP_Policies(self)
        self.schema = Dataset_Schema(self)
        self.tags = dmtg.Dataset_Tags(dataset=self)

    def display_url(self):
        return f'https://{self.domo_instance or self.full_auth.domo_instance}.domo.com/datasources/{self.id}/details/overview'

    @classmethod
    async def get_from_id(cls,
                          id: str,
                          full_auth: DomoFullAuth,
                          debug: bool = False, log_results: bool = False):

        # try:
        res = await dataset_routes.get_dataset_by_id(full_auth=full_auth,
                                                     id=id or cls.id, debug=debug)
        if res.status == 404:
            print(
                "f error retrieving get_from_id {full_auth.domo_instance} - {id} status = 404")
            raise ex.InvalidDataset(
                domo_instance=full_auth.domo_instance, dataset_id=id)

        # except Exception as e:
        #     print(e)
        #     return None

        if debug:
            pprint(res)

        dd = DictDot(res.response)
        ds = cls(
            full_auth=full_auth,
            id=dd.id,
            display_type=dd.displayType,
            data_provider_type=dd.dataProviderType,
            name=dd.name,
            description=dd.description,
            owner=dd.owner,
            formula=dd.properties.formulas.formulas,
            stream_id=dd.streamId,
            row_count=int(dd.rowCount),
            column_count=int(dd.columnCount)
        )

        if dd.tags:
            ds.tags.tag_ls = json.loads(dd.tags)

        if dd.certification:
            # print('class def certification', dd.certification)
            ds.certification = dmdc.DomoCertification._from_json(
                dd.certification)

        return ds

    @classmethod
    async def query_dataset(cls,
                            sql: str,
                            dataset_id: str,
                            dev_auth: DomoDeveloperAuth,
                            debug: bool = False,
                            session: aiohttp.ClientSession = None) -> pd.DataFrame:

        if debug:
            print("query dataset class method")
            print({'dataset_id': dataset_id,
                   'dev_auth': dev_auth})

        res = await dataset_routes.query_dataset_public(dev_auth=dev_auth, id=dataset_id, sql=sql, session=session,
                                                        debug=debug)

        if debug:
            print(res.response)

        if res.status == 200:
            df = pd.DataFrame(data=res.response.get('rows'),
                              columns=res.response.get('columns'))
            return df
        return None

    @classmethod
    async def query_dataset_private(cls,
                                    sql: str,
                                    dataset_id: str,
                                    full_auth: DomoFullAuth,
                                    debug: bool = False,
                                    session: aiohttp.ClientSession = None) -> pd.DataFrame:

        if debug:
            print("query dataset class method")
            print({'dataset_id': dataset_id,
                   'full_auth': full_auth})

        res = await dataset_routes.query_dataset_private(full_auth=full_auth, id=dataset_id, sql=sql, session=session,
                                                         debug=debug)

        return pd.DataFrame(res)

    async def upload_csv(self,
                         upload_df: pd.DataFrame = None,
                         upload_df_list: list[pd.DataFrame] = None,
                         upload_file: io.TextIOWrapper = None,

                         full_auth: DomoFullAuth = None,
                         upload_method: str = 'REPLACE',
                         dataset_id: str = None,
                         dataset_upload_id=None,
                         partition_key: str = None,
                         is_index: bool = True,
                         session: aiohttp.ClientSession = None,
                         debug: bool = False):

        full_auth = full_auth or self.full_auth
        dataset_id = dataset_id or self.id

        upload_df_list = upload_df_list or [upload_df]

        # stage 1 get uploadId
        if not dataset_upload_id:
            if debug:
                print(f"\n\n🎭 starting Stage 1")

            res = await dataset_routes.upload_dataset_stage_1(full_auth=full_auth,
                                                              dataset_id=dataset_id,
                                                              session=session,
                                                              data_tag=partition_key,
                                                              debug=debug
                                                              )
            if debug:
                print(f"\n\n🎭 Stage 1 response -- {res.status}")
                print(res)

            dataset_upload_id = res.response.get('uploadId')

        # stage 2 upload_dataset

        if debug:
            print(
                f"\n\n🎭 starting Stage 2 - {len(upload_df_list)} - number of parts")

        stage_2_res = None

        if upload_file:
            if debug:
                print('stage 2 - file')
            stage_2_res = await dataset_routes.upload_dataset_stage_2_file(full_auth=full_auth,
                                                                           dataset_id=dataset_id,
                                                                           upload_id=dataset_upload_id,
                                                                           part_id=1,
                                                                           file=upload_file,
                                                                           session=session, debug=debug)
            if debug:
                print(f"🎭 Stage 2 response -- {stage_2_res.status}")
                print(stage_2_res.print(is_pretty=True))

        else:
            if debug:
                print('stage 2 - df')
            stage_2_res = await asyncio.gather(*[dataset_routes.upload_dataset_stage_2_df(full_auth=full_auth,
                                                                                          dataset_id=dataset_id,
                                                                                          upload_id=dataset_upload_id,
                                                                                          part_id=index + 1,
                                                                                          upload_df=df,
                                                                                          session=session, debug=debug) for index, df in enumerate(upload_df_list)])

            if debug:
                for res in stage_2_res:
                    print(f"🎭 Stage 2 response -- {res.status}")
                    res.print(is_pretty=True)

        # return stage_2_res

#         # stage 3 commit_data
        if debug:
            print(f"\n\n🎭 starting Stage 3")
        await asyncio.sleep(10)

        stage3_res = await dataset_routes.upload_dataset_stage_3(full_auth=full_auth,
                                                                 dataset_id=dataset_id,
                                                                 upload_id=dataset_upload_id,
                                                                 update_method=upload_method,
                                                                 data_tag=partition_key,
                                                                 is_index=False,
                                                                 session=session,
                                                                 debug=debug)

        if debug:
            print(f"\n🎭 stage 3 res - {res.status}")
            print(stage3_res)

        if is_index:
            await self.index_dataset(full_auth=full_auth,
                                     dataset_id=dataset_id,
                                     debug=debug,
                                     session=session)

        return stage3_res

    async def index_dataset(self,
                            full_auth: DomoFullAuth = None,
                            dataset_id: str = None,
                            debug: bool = False,
                            session: aiohttp.ClientSession = None
                            ):

        full_auth = full_auth or self.full_auth
        dataset_id = dataset_id or self.id
        return await dataset_routes.index_dataset(full_auth=full_auth, dataset_id=dataset_id, debug=debug,
                                                  session=session)

    async def list_partitions(self,
                              full_auth: DomoFullAuth = None,
                              dataset_id: str = None,
                              debug: bool = False,
                              session: aiohttp.ClientSession = None
                              ):

        full_auth = full_auth or self.full_auth
        dataset_id = dataset_id or self.id

        res = await dataset_routes.list_partitions(full_auth=full_auth, dataset_id=dataset_id, debug=debug,
                                                   session=session)
        if res.status != 200:
            return None
        return res.response

    async def delete_partition(self,
                               dataset_partition_id: str,

                               dataset_id: str = None,
                               empty_df: pd.DataFrame = None,

                               full_auth: DomoFullAuth = None,

                               is_index: bool = True,
                               debug: bool = False,
                               session: aiohttp.ClientSession = None):

        is_close_session = True if not session else False

        session = session or aiohttp.ClientSession()
        full_auth = full_auth or self.full_auth
        dataset_id = dataset_id or self.id

#        if empty_df is None:
#            empty_df = await self.query_dataset_private(full_auth=full_auth,
#                                                        dataset_id=dataset_id,
#                                                        sql="SELECT * from table limit 1",
#                                                        debug=False)
#
#        await self.upload_csv(upload_df=empty_df.head(0),
#                              upload_method='REPLACE',
#                              is_index=is_index,
#                              partition_key=dataset_partition_id,
#                              session=session,
#                              debug=False)
        if debug:
            print(f"\n\n🎭 starting Stage 1")

        res = await dataset_routes.delete_partition_stage_1(full_auth=full_auth,
                                                            dataset_id=dataset_id,
                                                            dataset_partition_id=dataset_partition_id,
                                                            debug=debug, session=session)
        if debug:
            print(f"\n\n🎭 Stage 1 response -- {res.status}")
            print(res)

        stage_2_res = None
        if debug:
            print('starting Stage 2')
        stage_2_res = await dataset_routes.delete_partition_stage_2(full_auth=full_auth,
                                                                    dataset_id=dataset_id,
                                                                    dataset_partition_id=dataset_partition_id,
                                                                    debug=debug, session=session)
        if debug:
            print(f"\n\n🎭 Stage 2 response -- {stage_2_res.status}")

        stage_3_res = None
        if debug:
            print('starting Stage 3')
        stage_3_res = await dataset_routes.index_dataset(full_auth=full_auth,
                                                         dataset_id=dataset_id,
                                                         debug=debug, session=session)
        if debug:
            print(f"\n\n🎭 Stage 3 response -- {stage_3_res.status}")

        if is_close_session:
            await session.close()

        if debug:
            print(stage_3_res)

        if stage_3_res.status == 200:
            return res.response

    async def reset_dataset(self,
                            full_auth: DomoFullAuth = None,
                            is_index: bool = True,
                            debug: bool = False
                            ):
        execute_reset = input(
            "This function will delete all rows.  Type BLOW_ME_AWAY to execute:")

        if execute_reset != 'BLOW_ME_AWAY':
            print("You didn't type BLOW_ME_AWAY, moving on.")
            return None

        full_auth = full_auth or self.full_auth
        dataset_id = self.id

        if not full_auth:
            raise Exception("full_auth required")

        session = aiohttp.ClientSession()

        # create empty dataset to retain schema
        empty_df = await self.query_dataset_private(full_auth=full_auth,
                                                    dataset_id=dataset_id,
                                                    sql="SELECT * from table limit 1",
                                                    session=session,
                                                    debug=debug)
        empty_df = empty_df.head(0)

        # get partition list
#         partition_list = await dataset_routes.list_partitions(full_auth=full_auth,
#                                                               dataset_id=self.id,
#                                                               debug=debug,
#                                                               session=session)

#         if len(partition_list) > 0:
#             partition_list = chunk_list(partition_list, 100)

#             for index, pl in enumerate(partition_list):
#                 print(f'🥫 starting chunk {index + 1} of {len(partition_list)}')

#                 await asyncio.gather(*[self.delete_partition(full_auth=full_auth,
#                                                              dataset_partition_id=partition.get('partitionId'),
#                                                              session=session,
#                                                              empty_df=empty_df,
#                                                              debug=False) for partition in pl])
#                 if is_index:
#                     await self.index_dataset(session=session)

        res = await self.upload_csv(upload_df=empty_df,
                                    upload_method='REPLACE',
                                    is_index=is_index,
                                    session=session,
                                    debug=False)

        await session.close()
        return True

    async def delete(self,
                     dataset_id=None,
                     full_auth: DomoFullAuth = None,
                     debug: bool = False,
                     session: aiohttp.ClientSession = None):
        try:
            is_close_session = False

            if not session:
                session = aiohttp.ClientSession()
                is_close_session = True

            return await dataset_routes.delete(
                full_auth=full_auth or self.full_auth,
                dataset_id=dataset_id or self.id,
                debug=debug,
                session=session)

        finally:
            if is_close_session:
                await session.close()

    # async def create(self,
    #                   ds_name,
    #                   ds_type ='api',
    #                   schema = { "columns": [ {
    #                       "name": 'col1',
    #                       "type": 'LONG',
    #                       "metadata": None,
    #                       "upsertKey": False}
    #                   ]},
    #                   full_auth:DomoFullAuth = None,
    #                   debug:bool = False)



import datetime as dt
import importlib
from dataclasses import dataclass, field
from enum import Enum

import aiohttp
import Library.DomoClasses.DomoAuth as dmda
import Library.DomoClasses.routes.account_routes as account_routes
import Library.utils.DictDot as dcd

importlib.reload(account_routes)


@dataclass
class DomoAccount_Config_HighBandwidthConnector:
    aws_access_key: str
    aws_secret_key: str = field(repr = False)
    s3_staging_dir: str
    region: str = 'us-west-2'
    data_provider_type = 'amazon-athena-high-bandwidth'

    @classmethod
    def _from_json(cls, obj):

        dd = dcd.DictDot(obj)

        return cls(
            aws_access_key=dd.awsAccessKey,
            aws_secret_key=dd.awsSecretKey,
            s3_staging_dir=dd.s3StagingDir,
            region=dd.region
        )

    def to_json(self):
        return {"awsAccessKey": self.aws_access_key,
                "awsSecretKey": self.aws_secret_key,
                "s3StagingDir": self.s3_staging_dir,
                "region": self.region
                }


@dataclass
class DomoAccount_Config_AbstractCredential:
    credentials: dict

    @classmethod
    def _from_json(cls, obj):

        dd = dcd.DictDot(obj)

        return cls(
            credentials=dd.credentials,
        )

    def to_json(self):
        return {"credentials": self.credentials
                }


class AccountConfig(Enum):
    amazon_athena_high_bandwidth = {'type': 'amazon-athena-high-bandwidth',
                                    'config': DomoAccount_Config_HighBandwidthConnector}

    abstract_credential_store = {'type': 'abstract-credential-store',
                                 'config': DomoAccount_Config_AbstractCredential}


@dataclass
class DomoAccount:
    name: str
    data_provider_type: str
    id: int = None
    created_dt: dt.datetime = None
    modified_dt: dt.datetime = None
    full_auth: dmda.DomoFullAuth = field(repr=False, default=None)

    config:  AccountConfig = None

    @classmethod
    def _from_json(cls, obj: dict, full_auth: dmda.DomoFullAuth = None):
        import Library.utils.convert as cd

        dd = dcd.DictDot(obj)

        return cls(id=dd.id,
                   name=dd.displayName,
                   data_provider_type=dd.dataProviderType,
                   created_dt=cd.convert_epoch_millisecond_to_datetime(
                       dd.createdAt),
                   modified_dt=cd.convert_epoch_millisecond_to_datetime(
                       dd.modifiedAt),
                   full_auth=full_auth)

    @classmethod
    async def get_from_id(cls, full_auth: dmda.DomoFullAuth, account_id: int, session: aiohttp.ClientSession = None):
        res = await account_routes.get_account_from_id(full_auth=full_auth, account_id=account_id,
                                                       session=session
                                                       )

        import re

        if res.status != 200:
            return None

        obj = res.response
        acc = cls._from_json(obj, full_auth)

        res_config = await account_routes.get_account_config(full_auth=full_auth,
                                                             account_id=acc.id,
                                                             data_provider_type=acc.data_provider_type,
                                                             session=session)

        if res_config.status != 200:
            return acc

        enum_clean = re.sub("-", '_', acc.data_provider_type)

        account_config_names = [member.name for member in AccountConfig]

        if enum_clean not in account_config_names:
            return acc

        acc.config = AccountConfig[enum_clean].value.get('config')._from_json(
            res_config.response
        )

        return acc

    async def update_config(self,
                            full_auth: dmda.DomoFullAuth = None,
                            debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):

        full_auth = full_auth or self.full_auth

        # print(full_auth, self.id, self.data_provider_type, self.config.to_json())

        res = await account_routes.update_account_config(full_auth=full_auth,
                                                         account_id=self.id,
                                                         data_provider_type=self.data_provider_type,
                                                         config_body=self.config.to_json(),
                                                         debug=debug, log_results=log_results, session=session)

        obj = res.response
        acc = await self.get_from_id(full_auth=full_auth, account_id=obj.get('id'))
        acc.status = res.status
        acc.is_success = res.is_success

        return acc

    async def update_name(self,
                          account_name: str = None,
                          full_auth: dmda.DomoFullAuth = None,
                          debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):

        full_auth = full_auth or self.full_auth

        # print(full_auth, self.id, self.data_provider_type, self.config.to_json())

        res = await account_routes.update_account_name(full_auth=full_auth,
                                                       account_id=self.id,
                                                       account_name=account_name or self.name,
                                                       debug=debug, log_results=log_results, session=session)

        obj = res.response
        acc = await self.get_from_id(full_auth=full_auth, account_id=obj.get('id'))
        acc.status = res.status
        acc.is_success = res.is_success

        return acc

    def config_to_json(self):
        obj = {'displayName': self.name,
               'dataProviderType': self.data_provider_type,
               'name': self.data_provider_type,
               'configurations': self.config.to_json()}

        return obj

    async def create_account(self,
                             full_auth: dmda.DomoFullAuth = None,
                             #  config_body: dict,
                             debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):
        
        full_auth = full_auth or self.full_auth

        res = await account_routes.create_account(full_auth=full_auth,
                                                  config_body=self.config_to_json(),
                                                  debug=debug, log_results=log_results, session=session)

        if debug:
            print(res)
            
        if res.status != 200:
            return False

        obj = res.response
        acc = await self.get_from_id(full_auth=full_auth, account_id=obj.get('id'))
        acc.status = res.status
        acc.is_success = res.is_success
        return acc
    
    async def delete_account(self,
                             full_auth: dmda.DomoFullAuth = None,
                             #  config_body: dict,
                             debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):
        full_auth = full_auth or self.full_auth

        res = await account_routes.delete_account(full_auth=full_auth,
                                                  account_id = self.id,
                                                  debug=debug, log_results=log_results, session=session)

        if debug:
            print(res)
            
        if res.status != 200:
            return False

        return True
    
    async def share_account(self, 
                            full_auth : dmda.DomoFullAuth,
                            user_id: int,
                            is_v2: bool = False,
                            access_level: str = 'CAN_VIEW', # for v2 account_sharing beta CAN_VIEW, CAN_EDIT, CAN_SHARE
                            debug: bool = False,
                            log_results :bool = False,
                            session : aiohttp.ClientSession = None ):
        
        

        if is_v2:
            share_payload = account_routes.generate_share_account_payload_v2(user_id = user_id,
                                                                          access_level = access_level)
            return await account_routes.share_account_v2(full_auth = full_auth,
                                                  account_id = self.id,
                                                  share_payload = share_payload,
                                                  debug = debug, 
                                                  log_results = log_results,
                                                  session = session                                           
                                                 )
        
        share_payload = account_routes.generate_share_account_payload_v1(user_id = user_id)
        return await account_routes.share_account_v1(full_auth = full_auth,
                                                  account_id = self.id,
                                                  share_payload = share_payload,
                                                  debug = debug, 
                                                  log_results = log_results,
                                                  session = session                                           
                                                 )
        


# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/page.ipynb.

# %% auto 0
__all__ = ['get_page_by_id', 'get_page_definition', 'get_page_access_list', 'get_pages_adminsummary', 'update_page_layout',
           'put_writelock', 'delete_writelock']

# %% ../../nbs/routes/page.ipynb 2
import httpx
import pandas as pd

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda


# %% ../../nbs/routes/page.ipynb 3
async def get_page_by_id(auth: dmda.DomoAuth, page_id: str,
                         debug_api: bool = False, 
                         session: httpx.AsyncClient = None,
                         include_layout: bool = False
                         ) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/content/v3/stacks/{page_id}/cards'
    
    if include_layout:
        url+='?includeV4PageLayouts=true'
    
    res = await gd.get_data(
        auth=auth,
        url=url,
        method='GET',
        debug_api=debug_api,
        session = session,
    )

    res.response


    return res

# %% ../../nbs/routes/page.ipynb 6
async def get_page_definition(auth, page_id, debug_api: bool = False, session: httpx.AsyncClient = None):

    url = f"https://{auth.domo_instance}.domo.com/api/content/v3/stacks/{page_id}/cards"

    params = {"includeV4PageLayouts": "true",
              "parts": "metadata,datasources,library,drillPathURNs,certification,owners,dateInfo,subscriptions,slicers"}

    res = await gd.get_data(url,
                            method='GET',
                            auth=auth,
                            session=session,
                            params=params, debug_api=debug_api)

    return res


# %% ../../nbs/routes/page.ipynb 9
async def get_page_access_list(auth,
                              page_id,
                              is_expand_users: bool = True,
                              debug_api: bool = False, session: httpx.AsyncClient = None):
    """retrieves accesslist, which users and groups a page is shared with"""

    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/share/accesslist/page/{page_id}?expandUsers={is_expand_users}"

    res = await gd.get_data(url,
                            method='GET',
                            auth=auth,
                            session=session, debug_api=debug_api)

    # add group members to users response
    if is_expand_users:
        group_users = [user for group in res.response.get(
            'groups') for user in group.get('users')]
        users = res.response.get('users') + group_users
        res.response.update({'users': users})

    return res


# %% ../../nbs/routes/page.ipynb 12
async def get_pages_adminsummary(auth: dmda.DomoAuth,
                                 debug_loop: bool = False,
                                 debug_api : bool = False,
                                 limit = 35,
                                 session: httpx.AsyncClient = None):
    """retrieves all pages in instance user is able to see (but may not have been explicitly shared)"""

    url = f'https://{auth.domo_instance}.domo.com/api/content/v1/pages/adminsummary'

    offset_params = {
        'offset': 'skip',
        'limit': 'limit',
    }

    body = {"orderBy": "pageTitle", "ascending": True}

    def arr_fn(res) -> list[dict]:
        return res.response.get('pageAdminSummaries')

    res = await gd.looper(auth=auth,
                          method='POST',
                          url=url,
                          arr_fn=arr_fn,
                          offset_params=offset_params,
                          session=session,
                          loop_until_end=True,
                          body = body,
                          limit = limit, 
                          debug_loop=debug_loop, debug_api= debug_api)
    return res


# %% ../../nbs/routes/page.ipynb 14
async def update_page_layout(auth: dmda.DomoAuth, 
                            layout_id: str,
                             body: dict,
                         debug_api: bool = False):
    url = f'https://{auth.domo_instance}.domo.com/api/content/v4/pages/layouts/{layout_id}'


    res = await gd.get_data(
        auth=auth,
        url=url,
        body=body,
        method='PUT',
        debug_api=debug_api
    )

    if debug_api:
        print(res)

    return res

async def put_writelock(auth: dmda.DomoAuth, 
                        layout_id: str, 
                        user_id: str,
                        epoch_time: int,
                         debug_api: bool = False):
    url = f'https://{auth.domo_instance}.domo.com/api/content/v4/pages/layouts/{layout_id}/writelock'
    body= {
        "layoutId": layout_id,
        "lockHeartbeat": epoch_time,
        "lockTimestamp": epoch_time,
        "userId":user_id
    }
  
    res = await gd.get_data(
        auth=auth,
        url=url,
        body=body,
        method='PUT',
        debug_api=debug_api
    )

    if debug_api:
        print(res)

    return res

async def delete_writelock(auth: dmda.DomoAuth,
                           layout_id: str,
                         debug_api: bool = False):
    url = f'https://{auth.domo_instance}.domo.com/api/content/v4/pages/layouts/{layout_id}/writelock'
    res = await gd.get_data(
        auth=auth,
        url=url,
        method='DELETE',
        debug_api=debug_api
    )

    if debug_api:
        print(res)

    return res



import aiohttp

from ...utils.ResponseGetData import ResponseGetData


async def get_full_auth(domo_instance, domo_username, domo_password, session=None) -> ResponseGetData:
    is_close_session = False

    if not session:
        is_close_session = True
        session = aiohttp.ClientSession()

    url = f'https://{domo_instance}.domo.com/api/content/v2/authentication'

    tokenHeaders = {'Content-Type': 'application/json'}
    body = {'method': 'password', 'emailAddress': domo_username,
            'password': domo_password}

    res = await session.request(method='POST', url=url, headers=tokenHeaders, json=body)

    data = await res.json()

    if is_close_session:
        await session.close()

    if res.status == 200 and data.get('sessionToken'):
        return ResponseGetData(status=res.status,
                               is_success=True,
                               response=data)

    if not data.get('sessionToken'):
        return ResponseGetData(status=res.status,
                               is_success=False,
                               response=data)

    return gdr


async def get_developer_auth(domo_client_id, domo_client_secret, session=None) -> ResponseGetData:
    if not session:
        session = aiohttp.ClientSession(
            auth=aiohttp.BasicAuth(domo_client_id, domo_client_secret))

    url = f'https://api.domo.com/oauth/token?grant_type=client_credentials'

    res = await session.request(method='GET', url=url)
    data = await res.json()

    gdr = ResponseGetData(status=res.status,
                          is_success=True if res.status == 200 else False,
                          response=data)

    await session.close()
    return gdr



from typing import List

import aiohttp

from ...utils.ResponseGetData import ResponseGetData
from ..DomoAuth import DomoFullAuth
from .get_data import get_data


async def get_kpi_definition(full_auth: DomoFullAuth, card_id: str, debug: bool = False) -> ResponseGetData:
    url = f"https://{full_auth.domo_instance}.domo.com/api/content/v3/cards/kpi/definition"

    body = {"urn": card_id}

    res = await get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body=body,
        debug=False
    )

    return res


async def get_card_metadata(full_auth: DomoFullAuth, card_id: str, debug: bool = False) -> ResponseGetData:
    optional_params = "metadata,certification,datasources,owners,problems"
    url = f"https://{full_auth.domo_instance}.domo.com/api/content/v1/cards?urns={card_id}&parts={optional_params}"

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        debug=False
    )

    return res


def generate_body_search_cards_admin_summary(page_ids: List[str] = None,
                                             searchPages: bool = True,
                                             cardSearchText: str = '',
                                             pageSearchText: str = '') -> dict:
    body = {
        # "includeCardCertificationClause": false,
        # "cardCertification": [
        #     {
        #         "state": "CERTIFIED"
        #     }
        # ],
        "includeCardTitleClause": True if cardSearchText else False,
        "includePageTitleClause": True if pageSearchText else False,
        "notOnPage": False,
        "ascending": True,
        "orderBy": "cardTitle"
    }

    if cardSearchText:
        body.update({'cardTitleSearchText': cardSearchText})

    if pageSearchText:
        body.update({'pageTitleSearchText': pageSearchText})

    if page_ids:
        body.update({'pageIds': page_ids})

    return body


async def search_cards_admin_summary(full_auth: DomoFullAuth,
                                     body: dict,
                                     limit: int = 100,
                                     offset: int = 0,
                                     debug: bool = False,
                                     log_results: bool = False
                                     ) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/content/v2/cards/adminsummary?skip={offset}&limit={limit}'

    res = await get_data(
        auth=full_auth,
        url=url,
        body=body,
        method='POST',
        debug=debug,
        log_results=log_results
    )

    return res



from pprint import pprint

import aiohttp

from ...DomoClasses import DomoAuth as dmda
from ...utils import ResponseGetData as ResponseGetData
from . import get_data as gd


async def search_activity_log(full_auth: dmda.DomoFullAuth,
                              end_time: int,
                              start_time: int,
                              maximum: int, object_type=None,
                              session: aiohttp.ClientSession = None,
                              debug: bool = False):
    is_close_session = False

    if not session:
        session = aiohttp.ClientSession()
        is_close_session = True

    url = f'https://{full_auth.domo_instance}.domo.com/api/audit/v1/user-audits/{object_type if object_type else ""}'

    fixed_params = {
        'end': end_time,
        'start': start_time

    }

    offset_params = {
        'offset': 'offset',
        'limit': 'limit',
    }

    def arr_fn(res) -> list[dict]:
        return res.response

    res = await gd.looper(auth=full_auth,
                          method='GET',
                          url=url,
                          arr_fn=arr_fn,
                          fixed_params=fixed_params,
                          offset_params=offset_params,
                          session=session,
                          maximum=maximum, debug=debug)

    if is_close_session:
        await session.close()

    return res
